#pragma once
#ifndef APERTURE_HPP
#define APERTURE_HPP 1
#include <field.hpp>
#include <vector>
#include <source.hpp>

using namespace QOSM::Propagation;
using namespace QOSM::EM;

namespace QOSM::EM::Source{

    /**
     * @brief
     * Generic class modelling an aperture
     *
     */
    struct Aperture{
        /// Number of modes to be considered
        size_t numModes;

        /// [Hz] Frequency
        double freq;

        /// [1/m] wavenumber in the void
        double k0;

        /// Frame of the aperture
        Frame frame;

        /// Modal fields
        std::vector<FieldNx3> Ek, Hk;

        /// Modal impedences and admittences
        std::vector<cplx> Zk, Yk;
        
        /// Aperture side sampling
        nc::NdArray<double> xa, ya;

        /// Aperture side meshgrid
        nc::NdArray<double> Xa, Ya;

        /**
         * Instantiate an aperture
         * @param freq [Hz] Operative frequency.
         * @param numModes Number of modes to be considered.
         * @param frame Pose of the aperture w.r.t. the feeding source or it parent (i.e. a Horn).
         */
        Aperture(
            const double &freq = 0., 
            const size_t &numModes = 0., 
            const Frame &frame = Frame()
        );

        // TX
        /**
         * Compute electrical field radiated by the aperture.
         *
         * @param modes Modes coefficients that are propagative at the aperture.
         * @param r_pts_src Position vector where to compute the e-field.
         * @return the field at each position
         */
        virtual FieldNx3 eField(const std::vector<cplx> modes, const VectorNx3 &r) const{
            return FieldNx3(r.numRows());
        };

        /**
         * Compute magnetic field radiated by the aperture.
         *
         * @param modes Modes coefficients that are propagative at the aperture.
         * @param r_pts_src Position vector where to compute the h-field.
         * @return the field at each position
         */
        virtual FieldNx3 hField(const std::vector<cplx> modes, const VectorNx3 &r) const{
            return FieldNx3(r.numRows());
        };
          
        /**
         * Compute electric, magnetic fields and Poyinting vectors.
         *
         * @param r_pts_src Position vector where to compute the fields.
         * @return a tuple of electrical and magnetic fields and Poyinting vectors at each position
         */
        virtual std::tuple<FieldNx3, FieldNx3, VectorNx3> fields(
            const std::vector<cplx> modes, 
            const VectorNx3 &r
            ) const{
            return std::make_tuple(FieldNx3(r.numRows()), FieldNx3(r.numRows()), VectorNx3(r.numRows()));
        };

        // RX
        // compute RX from a source and position vectors => compute mode coefficients
        virtual nc::NdArray<cplx> computeRx(const std::shared_ptr<Source> &src, const VectorNx3 &r, const bool &onlyCentralPoint=false) const;
    };


    /**
     * @brief
     * Implementation of an Aperture for rectangular shape
     * 
     * The shape is modelled with its side lengths "Lx" and "Ly".
     * The aperture surface sampling is controlled by the parameter "numAperturePoints",
     * which is the number of points per axis (x and y).
     *
     */
    struct RectangularAperture : public Aperture{
        /// Number of points per axis (x and y)
        size_t numAperturePoints;

        /// [m] Aperture side length along x axis
        double Lx;

        /// [m] Aperture side length along y axis
        double Ly;

        /**
         * Instantiate a rectangular aperture
         * @param freq [Hz] Operative frequency.
         * @param Lx [m] width of the aperture.
         * @param Ly [m] height of the aperture.
         * @param numModes Number of modes to be considered.
         * @param numPtsAperturePerAxis Number of point per axis to sample the horn aperture surface.
         * @param frame Pose of the aperture w.r.t. the feeding source or it parent (i.e. a Horn).
         */
        RectangularAperture(
            const double &freq, 
            const double &Lx, 
            const double &Ly, 
            const size_t &numModes, 
            const size_t &numPtsAperturePerAxis, 
            const Frame &frame = Frame()
        );

        // TX
        /**
         * Compute electrical field radiated by the aperture.
         *
         * @param modes Modes coefficients that are propagative at the aperture.
         * @param r_pts_src Position vector where to compute the e-field.
         * @return the field at each position
         */
        FieldNx3 eField(const std::vector<cplx> modes, const VectorNx3& r) const;

        /**
         * Compute magnetic field radiated by the aperture.
         *
         * @param modes Modes coefficients that are propagative at the aperture.
         * @param r_pts_src Position vector where to compute the h-field.
         * @return the field at each position
         */
        FieldNx3 hField(const std::vector<cplx> modes, const VectorNx3& r) const;

        /**
         * Compute electric, magnetic fields and Poyinting vectors.
         *
         * @param r_pts_src Position vector where to compute the fields.
         * @return a tuple of electrical and magnetic fields and Poyinting vectors at each position
         */
        std::tuple<FieldNx3, FieldNx3, VectorNx3> fields(
            const std::vector<cplx> modes, 
            const VectorNx3 &r
        ) const;
        
        /**
         * Compute eletrical fields received from the source at the aperture at each position vector.
         *
         * @param src Source of radiation.
         * @param r_pts_src Position vector where to compute the fields.
         * @param onlyCentralPoint Set to true for assuing the field is almost uniform at the aperture.
         * @return the electrical field at aperture sample points for
         */
        FieldNx3 eFieldRx(
            const std::shared_ptr<Source>& src, 
            const VectorNx3& r, 
            const bool& onlyCentralPoint=false
        ) const;

        /**
         * Perform modal expantion of the eletrical field received from the source at the aperture at each position vector.
         *
         * @param src Source of radiation.
         * @param r_pts_src Position vector where to compute the fields.
         * @param onlyCentralPoint Set to true for assuing the field is almost uniform at the aperture.
         * @return the the modes coefficients for each position
         */
        // RX
        nc::NdArray<cplx> computeRx(
            const std::shared_ptr<Source>& src, 
            const VectorNx3& r_pts_src, 
            const bool& onlyCentralPoint=false
        ) const;

        /**
         * Perform modal expantion of the eletrical field received at the aperture at each position vector.
         * @see FieldNx3::eFieldRx(const std::shared_ptr<Source>&, const VectorNx3&, const bool&)
         *
         * @param fields Electrical fields at each sample point of the aperture for each position.
         * @return the the modes coefficients for each position
         */
        nc::NdArray<cplx> computeRx(const FieldNx3& fields) const;

        /**
         * @return a vector of flag whether each mode is propagative or not
         */
        std::vector<bool> propagativeModes() const;
    };
    



    /**
     * @brief
     * Implementation of an Aperture for circular shape
     * 
     * The shape is modelled with its radius length "radius".
     * The aperture surface sampling is controlled by the parameter "numAperturePoints",
     * which is the number of points per axis (rho and theta).
     *
     */
    struct CircularAperture : public Aperture{
        /// Number of points per axis (x and y)
        std::array<size_t, 2> numAperturePoints;

        /// [m] Radius of the aperture
        double radius;

        /**
         * Instantiate a rectangular aperture
         * @param freq [Hz] Operative frequency.
         * @param radius [m] Radius of the aperture.
         * @param numModes Number of modes to be considered.  // only TE11 implemented so far
         * @param numPtsAperturePerAxis Number of point per axis (rho, theta) to sample the horn aperture surface.
         * @param frame Pose of the aperture w.r.t. the feeding source or it parent (i.e. a Horn).
         * 
         * @todo All TE mode (only TE11 so far)
         */
        CircularAperture(
            const double &freq, 
            const double &radius, 
            const size_t &numModes, 
            const std::array<size_t, 2> &numPtsAperturePerAxis, 
            const Frame &frame = Frame()
        );

        // TX
        /**
         * Compute electrical field radiated by the aperture.
         *
         * @param modes Modes coefficients that are propagative at the aperture.
         * @param r_pts_src Position vector where to compute the e-field.
         * @return the field at each position
         */
        FieldNx3 eField(const std::vector<cplx> modes, const VectorNx3& r) const;

        /**
         * Compute magnetic field radiated by the aperture.
         *
         * @param modes Modes coefficients that are propagative at the aperture.
         * @param r_pts_src Position vector where to compute the h-field.
         * @return the field at each position
         */
        FieldNx3 hField(const std::vector<cplx> modes, const VectorNx3& r) const;

        /**
         * Compute electric, magnetic fields and Poyinting vectors.
         *
         * @param r_pts_src Position vector where to compute the fields.
         * @return a tuple of electrical and magnetic fields and Poyinting vectors at each position
         */
        std::tuple<FieldNx3, FieldNx3, VectorNx3> fields(
            const std::vector<cplx> modes, 
            const VectorNx3 &r
        ) const;
        
        /**
         * Compute eletrical fields received from the source at the aperture at each position vector.
         *
         * @param src Source of radiation.
         * @param r_pts_src Position vector where to compute the fields.
         * @param onlyCentralPoint Set to true for assuing the field is almost uniform at the aperture.
         * @return the electrical field at aperture sample points for
         */
        FieldNx3 eFieldRx(
            const std::shared_ptr<Source>& src, 
            const VectorNx3& r, 
            const bool& onlyCentralPoint=false
        ) const;

        /**
         * Perform modal expantion of the eletrical field received from the source at the aperture at each position vector.
         *
         * @param src Source of radiation.
         * @param r_pts_src Position vector where to compute the fields.
         * @param onlyCentralPoint Set to true for assuing the field is almost uniform at the aperture.
         * @return the the modes coefficients for each position
         */
        // RX
        nc::NdArray<cplx> computeRx(
            const std::shared_ptr<Source>& src, 
            const VectorNx3& r_pts_src, 
            const bool& onlyCentralPoint=false
        ) const;

        /**
         * Perform modal expantion of the eletrical field received at the aperture at each position vector.
         * @see FieldNx3::eFieldRx(const std::shared_ptr<Source>&, const VectorNx3&, const bool&)
         *
         * @param fields Electrical fields at each sample point of the aperture for each position.
         * @return the the modes coefficients for each position
         */
        nc::NdArray<cplx> computeRx(const FieldNx3& fields) const;

        /**
         * @return a vector of flag whether each mode is propagative or not
         */
        std::vector<bool> propagativeModes() const;
    };
}
#endif