#pragma once
#ifndef BVH_HPP
#define BVH_HPP 1
#include <stdexcept> // For std::runtime_error
#include <memory> // For std::shared_ptr
#include <cmath>
#include <vector>
#include <triangle.hpp>

using namespace QOSM::Pose;
using namespace QOSM::Propagation;
using namespace QOSM::Item;

namespace QOSM::Item{
    struct Intersection{
        triangle_ptr triangle;
        double t, u, v;
        bool valid;
    };

    class BVHNode {
    public:
        Vec3 bounding_box[2]; // min and max corners of the bounding box
        std::shared_ptr<BVHNode> left;
        std::shared_ptr<BVHNode> right;
        std::vector<triangle_ptr> triangles;

        /**
         * Create a node.
         *
         * @param bounding_box Min and max corner position vectors, representing the bounding volume that encloses all the primitives or child nodes contained within it.
         * @param left Child node on the left of the split axis.
         * @param right Child node on the right of the split axis.
         * @param triangles Triangles located in the bounding box.
         * @return The created BVH node.
         */
        BVHNode(Vec3 bb[2], std::shared_ptr<BVHNode> left_ptr = nullptr, std::shared_ptr<BVHNode> right_ptr = nullptr, std::vector<triangle_ptr> tri = {});

        bool bounding_box_intersect(const QOSM::Propagation::Ray& ray);
    };

    /**
     * Build the BVH recursively.
     *
     * @param triangles All triangles to be considered.
     * @param depth Current depth of the BVH in construction.
     * @return The root BVH Node, nullptr if no triangles have been found.
     */
    std::shared_ptr<BVHNode> build_bvh(const std::vector<triangle_ptr>& triangles, int depth = 0);

    /**
     * Traverse the BVH to find an intersection with the ray.
     *
     * @param ray The ray as a tuple (origin, direction).
     * @param bvh_node The BVH Node potentially traversed by the ray. If it is a leaf, a precise intersection test is done with its triangles. If it is an internal node, a recursive call with its children will be carried out.
     * @return An Intersection (Triangle, distance, u, v, valid) 
     */
    Intersection traverse_bvh_policy0(const QOSM::Propagation::Ray& ray, const std::shared_ptr<BVHNode> &bvh_node);

    /**
     * Traverse the BVH to find potential intersections with the ray.
     *
     * @param ray The ray as a tuple (origin, direction).
     * @param bvh_node The BVH Node potentially traversed by the ray. If it is a leaf, a precise intersection test is done with its triangles. If it is an internal node, a recursive call with its children will be carried out.
     * @return A vector containing each intersection found (Triangle, distance, u, v, valid).
     */
    std::vector<Intersection> traverse_bvh_policy1(const QOSM::Propagation::Ray& ray, const std::shared_ptr<BVHNode> &bvh_node);

    /**
     * Find a potential intersection between the ray and the BVH.
     *
     * @param ray The ray as a tuple (origin, direction).
     * @param bvh_root The BVH root node, describing the entire geometry of the scene to be considered.
     * @param policy If 0, return the first intersection found, assuming it is necessarily the closest one. If 1, return all intersections found and choose the closest one.
     * @return An Intersection (Triangle, distance, u, v, valid) 
     */
    Intersection ray_intersect_bvh(const QOSM::Propagation::Ray& ray, const std::shared_ptr<BVHNode> &bvh_root, int policy = 0);

} // namespace
#endif