#pragma once
#ifndef GRID_HPP
#define GRID_HPP 1
#include <pose.hpp>
#include <map>

using namespace QOSM;
using namespace QOSM::Pose;

namespace QOSM::Tools{

    enum class PlaneType{
        XY = 0,
        XZ,
        YX,
        YZ,
        ZX,
        ZY
    };

    /**
     * @class Grid
     * @brief Represents a 2D grid of points in a specified plane.
     *
     * The Grid class defines a structured 2D set of points within given parameter 
     * ranges. It supports various grid generation methods and transformations such 
     * as rotation and scaling.
     */
    class Grid{
        protected:
            /**
             * @brief Creates grid points within the specified parameter ranges.
             * @param uRange The range of values for the u-coordinate (min, max).
             * @param vRange The range of values for the v-coordinate (min, max).
             * @param numPoints Number of points along u and v directions.
             * @param n A scaling factor.
             * @param plane The plane type in which the grid is defined.
             */
            void createPoints(const std::array<double,2> &uRange, const std::array<double,2> &vRange, const std::array<int,2> &numPoints,  const double &n, const PlaneType &plane);
        
        public:
            std::array<unsigned int, 2> shape{0, 0}; ///< Grid dimensions (rows, cols).
            nc::NdArray<double> u; ///< U-coordinates of grid points.
            nc::NdArray<double> v; ///< V-coordinates of grid points.
            VectorNx3 points; ///< 3D coordinates of grid points.
            PlaneType type; ///< Type of plane where the grid is defined.

            /**
             * @brief Constructs a Grid with adaptive step size.
             * @param uRange Range and step for the u-coordinate (min, max, step).
             * @param vRange Range and step for the v-coordinate (min, max, step).
             * @param n A scaling factor.
             * @param plane The plane type in which the grid is defined.
             */
            Grid(const std::array<double,3>& uRange, 
                 const std::array<double,3>& vRange, 
                 const double& n, 
                 const PlaneType& plane);
                
            /**
             * @brief Constructs a Grid with a fixed number of points.
             * @param uRange Range of the u-coordinate (min, max).
             * @param vRange Range of the v-coordinate (min, max).
             * @param numPoints Number of points along u and v directions.
             * @param n A scaling factor.
             * @param plane The plane type in which the grid is defined.
             */
            Grid(const std::array<double,2> &uRange, 
                 const std::array<double,2> &vRange, 
                 const std::array<int,2> &numPoints,  
                 const double &n, 
                 const PlaneType &plane);
            
            /**
             * @brief Rotates the grid around the U-axis.
             * @param centre The center of rotation.
             * @param angle The rotation angle (in radians by default).
             * @param deg If true, the angle is interpreted in degrees.
             */
            void rotateAlongU(const Vec3 &centre, const double &angle, const bool &deg=false);
            
            /**
             * @brief Rotates the grid around the V-axis.
             * @param centre The center of rotation.
             * @param angle The rotation angle (in radians by default).
             * @param deg If true, the angle is interpreted in degrees.
             */
            void rotateAlongV(const Vec3 &centre, const double &angle, const bool &deg=false);
            
            /**
             * @brief Rotates the grid using a quaternion.
             * @param centre The center of rotation.
             * @param rot The quaternion representing the rotation.
             */
            void rotate(const Vec3 &centre, const Quaternion &rot);
            
            /**
             * @brief Scales the grid.
             * @param scale The scaling factor.
             */
            void operator*=(const double &scale);
            
            /**
             * @brief Uniformly scales down the grid.
             * @param scale The scaling factor.
             */
            void operator/=(const double &scale);
            
            /**
             * @brief Translates the grid by a given offset.
             * @param offset The translation vector.
             */
            void operator+=(const Vec3 &offset);
    
            /**
             * @brief Translates the grid in the opposite direction of a given offset.
             * @param offset The translation vector.
             */
            void operator-=(const Vec3 &offset);
    };
}
#endif