#pragma once
#ifndef PROPAGATION_HPP
#define PROPAGATION_HPP
#include <deque>
#include <vector>
#include <beam.hpp>
#include <triangle.hpp>
#include <field.hpp>
#include <vsrc.hpp>

namespace qpropa = QOSM::Propagation;
namespace qitem = QOSM::Item;
namespace qem = QOSM::EM;
namespace qpose = QOSM::Pose;
namespace qsrc = QOSM::EM::Source;

/**
 * @brief Expands an incoming field into a set of vectorial Gaussian beams (2D case).
 *
 * @param E [complex Nx3] Incoming electric field at each point of the expansion surface.
 * @param S [double Nx3] Incoming Poynting vector at each point of the expansion surface.
 * @param src_r_pts [double Nx3] Position vectors of the expansion surface's points.
 * @param freq [Hz] Operating frequency.
 * @param dim Dimension of the expansion (2D case).
 * @param w0 Beam waist parameter.
 * @param pw_threshold Minimum power ratio (relative to the maximum) required for a point to be considered.
 * @return A vector of Gaussian beams representing the field expansion.
 */
std::vector<qpropa::Beam> fieldExpansion(
    const FieldNx1 &E, 
    const VectorNx3 &S, 
    const VectorNx3 &src_r_pts,
    const double &freq, 
    const int& dim,
    const double &w0, 
    const double pw_threshold = 1e-5
);

/**
 * @brief Expands an incoming field into a set of vectorial Gaussian beams (3D case).
 *
 * @param E [complex Nx3] Incoming electric field at each point of the expansion surface.
 * @param S [double Nx3] Incoming Poynting vector at each point of the expansion surface.
 * @param src_r_pts [double Nx3] Position vectors of the expansion surface's points.
 * @param freq [Hz] Operating frequency.
 * @param w0 Beam waist parameter.
 * @param pw_threshold Minimum power ratio (relative to the maximum) required for a point to be considered.
 * @return A vector of Gaussian beams representing the field expansion.
 */
std::vector<qpropa::Beam> fieldExpansion(
    const FieldNx3 &E, 
    const VectorNx3 &S, 
    const VectorNx3 &src_r_pts,
    const double &freq, 
    const double &w0, 
    const double pw_threshold = 1e-5
);

/**
 * @brief Performs beam tracing through a quasi-optical system.
 *
 * @param geometry [Nx1 Triangles] List of triangles representing the system geometry.
 * @param beams [Mx1 Beams] Beams representing the field to be propagated.
 * @param numBouces [2 integers] Maximum allowed bounces for (reflection, transmission).
 * @param pw_threshold Minimum power ratio (relative to incident beam power) for generated beams to be considered.
 * @param incrementBouncing Flag indicating whether to increment the bounce count.
 * @return A deque of beams after propagation.
 */
std::deque<qpropa::Beam> beamTracing(
    std::vector<std::shared_ptr<qitem::Triangle>>& geometry, 
    std::vector<qpropa::Beam>& beams, 
    const std::array<int, 2> &numBouces, 
    const double &pw_threshold, 
    const bool &incrementBouncing
);

/**
 * @brief Performs beam tracing through a quasi-optical system.
 *
 * @param geometry [Nx1 Triangles] List of triangles representing the system geometry.
 * @param beams [Mx1 Beams] Beams representing the field to be propagated.
 * @param numBouces [2x1 integers] Maximum allowed bounces for (reflection, transmission).
 * @param thresh Minimum power ratio (relative to incident beam power) for generated beams to be considered.
 * @param incrementBouncing Flag indicating whether to increment the bounce count.
 * @return A deque of beams after propagation.
 */
std::deque<qpropa::Beam> beamTracing(
    std::vector<qitem::Triangle>& geometry, 
    std::vector<qpropa::Beam>& beams, 
    const std::array<int, 2> &numBouces, 
    const double &thresh, 
    const bool &incrementBouncing
);

/**
 * @brief Performs beam tracing through a single surface.
 *
 * @param normals [Nx3 doubles] Normal vectors to the surface at each point.
 * @param curvatures [Nx1 doubles] Surface curvature at each point.
 * @param dioptre Optical interface for beam-surface interactions.
 * @param beams [Nx1 Beams] Beams representing the field to be propagated.
 * @param thresh Minimum power ratio (relative to incident beam power) for generated beams to be considered.
 * @return A deque of beams after interaction with the surface.
 */
std::deque<qpropa::Beam> surfaceBeamTracing(
    std::vector<qpose::Vec3>& normals, 
    std::vector<double>& curvatures, 
    qitem::Dioptre &dioptre, 
    std::vector<qpropa::Beam>& beams, 
    const double &thresh
);

/**
 * @brief Computes the collected power through an aperture.
 *
 * @param src Electromagnetic source.
 * @param aperturePoints [Nx3] Positions of points on the aperture surface in the source frame.
 * @param dudv Product of du and dv, the distance between points along each axis.
 * @param z_probe_src [1x3] Probe direction in the source frame.
 * @return The collected power through the aperture.
 */
double computeCollectedPower(
    const std::shared_ptr<qsrc::Source> &src, 
    const qpose::VectorNx3 &aperturePoints, 
    const double &dudv, 
    const qpose::Vec3 &z_probe_src
);

#endif