#ifndef TRIANGLE_hpp
#define TRIANGLE_hpp
#include "ray.hpp"

using namespace QOSM::Pose;

namespace QOSM::Item{

    struct Medium{
        /// Real part of the complex index of refraction
        double n;

        /// Imaginary part of the complex index of refraction
        double kappa;

        /// Complex wave impedence of the medium
        std::complex<double> Z_;
        
        /// Flag defining if the medium has a metallic surface
        bool isMetallic=false;
        
        /// Flag defining if the medium has a Perfec Electric Conductor surface
        bool isPEC=false;

        /**
         * Create a medium as Free Space (ior=1.)
         * @param isPEC Flag defining if the medium is a Perfect Electric Conductor (if True, then no need of IOR)
         */
        Medium(const bool isPEC=false);

        /**
         * Create a medium with its complex index of refraction
         * @param ior The complex index of refraction as: \f$ ior = n - j\kappa \f$, with \f$\kappa \geq 0\f$
         * @param isMetallic Flag defining if the medium is a a metallic surface
         */
        Medium(const std::complex<double> &ior_, const bool isMetallic=false);

        /**
         * Create a medium with its complex index of refraction
         * @param ior The complex index of refraction as a tuple : \f$(n, \kappa)\f$, using the convention \f$ ior = n - j\kappa \f$, with \f$\kappa \geq 0\f$
         * @param isMetallic Flag defining if the medium has a metallic surface
         */
        Medium(const std::pair<double, double> &ior, const bool isMetallic=false);

        /**
         * Set the medium parameter using the complex permitivity \f$ \epsilon_r = \epsilon_r' - j\epsilon_r" \f$
         * @param epsilon_r The complex permitivity
         */
        void setWithPermittivity(const std::complex<double> &epsilon_r_);

        /**
         * Set the medium parameter using the complex permitivity \f$ \epsilon_r = \epsilon_r' - j\epsilon_r" \f$
         * @param epsr_re Real part of the permitivity \f$ \epsilon_r' \f$
         * @param epsr_im Imaginary part of the permitivity \f$ \epsilon_r" \f$
         */
        void setWithPermittivity(const double &epsr_re, const double &epsr_imag);

        /**
         * Set the medium parameter using the relative permitivity and the conductivity
         * @param epsr_re Real part of the permitivity
         * @param conductivity [S/m] Conductivity of the medium
         * @param freq [Hz] Frequence at which the conductivity is provided
         */
        void setWithConductivity(const double &epsr_re, const double &conductivity, const double &freq);

        /**
         * Retrieve the complex index of refraction
         * @return a pair of double as <real part, imaginary part> of the index of refraction
         */
        std::pair<double, double> get_ior();
    };

    struct Dioptre{
        /**
         * Create a dioptre with a either PEC (Perfect Electric Conductor) surface or no defined media
         */
        Dioptre(bool is_pec = false);

        /**
         * Create a  dioptre with the two media
         * @param external_medium Refractive medium
         * @param internal_medium Incident medium
         */
        Dioptre(const Medium &external_medium, const Medium &internal_medium);

        /**
         * Create a  dioptre with the two media defined with their complex index of refraction and material properties
         * @param ior1 Complex IOR of Incident medium as a tuple : \f$(n, \kappa)\f$, using the convention \f$ ior = n - j\kappa \f$, with \f$\kappa \geq 0\f$
         * @param ior2 Complex IOR of Refractive medium as a tuple : \f$(n, \kappa)\f$, using the convention \f$ ior = n - j\kappa \f$, with \f$\kappa \geq 0\f$
         * @param isMetallic Flag to set the dioptre surface has a Metallic surface
         */
        Dioptre(const std::pair<double, double> &ior1, const std::pair<double, double> &ior2, const bool isMetallic=false);

        /// @brief Real part of the complex indexes of refraction
        double n1, n2;
        
        /// @brief Imaginary part of the complex indexes of refraction
        double kappa1, kappa2;

        /// Flag for Metallic surface. If true, no transmission will be accounted inside the metal 
        bool isMetallic = false;
        
        /// Flag for PEC surface. If true, no need for both side because total reflection is assumed
        bool isPEC = false;
    };

    class Triangle{
        public:
            Vec3 v0, v1, v2, n0, n1, n2, edge1, edge2, n, b;
            double c0, c1, c2;
            double c;
            Dioptre dioptre;
            int idTriangle;
            std::array<int, 2> bounces;

            /**
             * Create a Triangle using 3 points (clockwise order)
             * @param vertices Array of 3 points (3D vectors), one per corner 
             * @param normals Array of 3 normal vector (3D vectors), defined at each corner
             * @param edges Array of 2 edges (3D vectors), defined form the first corner
             * @param curvature Array of 2 double, defining the surface's curvature at each corner
             * @param dioptre Dioptre defined for the triangle surface
             * @param idTriangle Unique triangle identifier
             */
            Triangle(const std::array<Vec3, 3> &vertices, 
                     const std::array<Vec3, 3> &normals,
                     const std::array<Vec3, 2> &edges, 
                     const std::array<double, 3> &curvatures, 
                     const Dioptre &dioptre,
                     const int &idTriangle);

            /**
             * Check if a ray intersects the triangle.
             *
             * @param ray The ray as a tuple (origin, direction).
             * @return true if an intersection has been found, false otherwise
             */
            bool intersect(const QOSM::Propagation::Ray &ray, double &t, double &u, double &v) const;

            /**
             * Interpolate the normal for a point inside the triangle.
             *
             * @param curvature The interpolated curvature (result)
             * @param u u-coord in barycentric coordinates
             * @param v v-coord in barycentric coordinates
             * @return The interpolated normal as a 3D vector
             */
            Vec3 normal(double &u, double &v) const;

            /**
             * Interpolate the curvature for a point inside the triangle.
             *
             * @param normal The interpolated normal (result)
             * @param u u-coord in barycentric coordinates
             * @param v v-coord in barycentric coordinates
             * @return The interpolated curvature
             */
            double curvature(double &u, double &v) const;
    };
    using triangle_ptr  = std::shared_ptr<QOSM::Item::Triangle>;
}
#endif